DROP TABLE call_parts;
DROP TABLE calls_merged;

DROP TABLE objects;
DROP TABLE objects_deduped;

DROP TABLE tables;
DROP TABLE tables_deduped;

DROP TABLE table_rows;
DROP TABLE table_rows_deduped;

DROP TABLE files;
DROP TABLE files_deduped;
